/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    MemoryX::ArmarXObjects::GraphImportExport
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "GraphImportExport.h"

#include <armarx/navigation/location/aron/Location.aron.generated.h>
#include <armarx/navigation/location/constants.h>
#include <armarx/navigation/graph/aron/Graph.aron.generated.h>
#include <armarx/navigation/graph/constants.h>

#include <MemoryX/libraries/memorytypes/MemoryXTypesObjectFactories.h>
#include <MemoryX/core/MemoryXCoreObjectFactories.h>

#include <RobotAPI/libraries/core/FramedPose.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <iomanip>


namespace armarx::nav
{

    GraphImportExport::GraphImportExport()
    {
        properties.memoryName = "Navigation";
        properties.locationCoreSegmentName = loc::coreSegmentName;
        properties.graphCoreSegmentName = graph::coreSegmentName;
    }

    armarx::PropertyDefinitionsPtr GraphImportExport::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr def = new ComponentPropertyDefinitions(getConfigIdentifier());

        // Publish to a topic (passing the TopicListenerPrx).
        // def->topic(myTopicListener);

        // Subscribe to a topic (passing the topic name).
        // def->topic<PlatformUnitListener>("MyTopic");

        // Use (and depend on) another component (passing the ComponentInterfacePrx).
        def->component(proxies.priorKnowledge);
        def->component(proxies.graphNodePoseResolver, "GraphNodePoseResolver");


        // Add a required property. (The component won't start without a value being set.)
        // def->required(properties.boxLayerName, "p.box.LayerName", "Name of the box layer in ArViz.");

        // Add an optionalproperty.
        // def->optional(properties.boxLayerName, "p.box.LayerName", "Name of the box layer in ArViz.");


        return def;
    }


    void GraphImportExport::onInitComponent()
    {
        // Topics and properties defined above are automagically registered.

        // Keep debug observer data until calling `sendDebugObserverBatch()`.
        setDebugObserverBatchModeEnabled(true);
    }


    void GraphImportExport::onConnectComponent()
    {
        // Get proxies.
        if (proxies.priorKnowledge->hasGraphSegment())
        {
            proxies.graphSegment = proxies.priorKnowledge->getGraphSegment();
        }
        proxies.navigationWriter = memoryNameSystem.useWriter(armem::MemoryID(properties.memoryName));

        // Setup and start the remote GUI.
        refreshScenes();
        RemoteGui_startRunningTask();
    }


    void GraphImportExport::onDisconnectComponent()
    {
    }


    void GraphImportExport::onExitComponent()
    {
    }


    std::string GraphImportExport::getDefaultName() const
    {
        return "GraphImportExport";
    }


    void GraphImportExport::createRemoteGuiTab(const std::vector<std::string>& sceneNames)
    {
        using namespace armarx::RemoteGui::Client;

        tab.sceneComboBox.setOptions(sceneNames);
        tab.sceneRefreshButton.setLabel("Refresh");

        tab.providerSegmentLine.setValue(getName());

        tab.dryRun.setValue(false);

        tab.locationsMemoryXToArMemButton.setLabel("Locations MemoryX -> ArMem");
        tab.locationsArMemToMemoryXButton.setLabel("Locations ArMem -> MemoryX (WIP)");
        tab.locationsClearArMemButton.setLabel("Clear ArMem Locations");

        tab.graphMemoryXToArMemButton.setLabel("Graph MemoryX -> ArMem (WIP)");
        tab.graphArMemToMemoryXButton.setLabel("Graph ArMem -> MemoryX (WIP)");
        tab.graphClearArMemButton.setLabel("Clear ArMem Graphs");


        GridLayout grid;
        int row = 0;
        {
            grid.add(Label("Scene:"), {row, 0}).add(tab.sceneComboBox, {row, 1}).add(tab.sceneRefreshButton, {row, 2});
            ++row;

            grid.add(Label("Provider Segment:"), {row, 0}).add(tab.providerSegmentLine, {row, 1});
            ++row;

            grid.add(Label("Dry Run:"), {row, 0}).add(tab.dryRun, {row, 1});
            ++row;

            grid.add(tab.locationsMemoryXToArMemButton, {row, 0}).add(tab.locationsArMemToMemoryXButton, {row, 1})
                    .add(tab.locationsClearArMemButton, {row, 2});
            ++row;

            grid.add(tab.graphMemoryXToArMemButton, {row, 0}).add(tab.graphArMemToMemoryXButton, {row, 1})
                    .add(tab.graphClearArMemButton, {row, 2});
            ++row;
        }

        VBoxLayout root = {grid, VSpacer()};
        RemoteGui_createTab(getName(), root, &tab);
    }


    void GraphImportExport::RemoteGui_update()
    {
        if (tab.sceneRefreshButton.wasClicked())
        {
            refreshScenes();
        }
        if (tab.locationsMemoryXToArMemButton.wasClicked())
        {
            exportLocations(tab.sceneComboBox.getValue());
        }

        if (tab.locationsMemoryXToArMemButton.wasClicked())
        {
            clearArMemProviderSegment(getLocationProviderSegmentID());
        }
        if (tab.graphMemoryXToArMemButton.wasClicked())
        {
            clearArMemProviderSegment(getGraphProviderSegmentID());
        }
    }


    void GraphImportExport::refreshScenes()
    {
        const ::Ice::StringSeq sceneNames = proxies.graphSegment->getScenes();
        createRemoteGuiTab(sceneNames);
    }


    void GraphImportExport::clearArMemProviderSegment(const armem::MemoryID& providerSegmentID)
    {
        const bool clearWhenExists = true;
        auto result = proxies.navigationWriter.addSegment(providerSegmentID, clearWhenExists);
        if (result.success)
        {
            ARMARX_IMPORTANT << "Cleared ArMem provider segment " << providerSegmentID << ".";
        }
        else
        {
            ARMARX_WARNING << result.errorMessage;
        }
    }


    void GraphImportExport::exportLocations(const std::string& sceneName)
    {
        const armem::Time time = armem::Time::now();
        armem::Commit commit;

        memoryx::GraphNodeBaseList graphNodes = proxies.graphSegment->getNodesByScene(sceneName);
        for (memoryx::GraphNodeBasePtr& node : graphNodes)
        {
            ARMARX_CHECK_NOT_NULL(node);
            if (not node->isMetaEntity())
            {
                // ID is just some random MongoDB hash
                const std::string nodeId = node->getId();
                // This is the readable name entered in the GUI.
                const std::string name = node->getName();

                armarx::FramedPosePtr pose = armarx::FramedPosePtr::dynamicCast(node->getPose());
                ARMARX_CHECK_NOT_NULL(pose);

                FramedPosePtr globalNodePose = FramedPosePtr::dynamicCast(proxies.graphNodePoseResolver->resolveToGlobalPose(node));
                ARMARX_CHECK_NOT_NULL(globalNodePose);

                // `pose` and `globalNodePose` seem to be identical. Is the last step necessary?
                // Maybe `pose` could be non-global.

                ARMARX_VERBOSE
                        << std::setprecision(2) << std::fixed
                        << "Processing node " << (commit.updates.size() + 1)
                        << "\n- ID: \t" << nodeId
                        << "\n- Name: \t" << name
                        << "\n- Pose: \n" << pose->toEigen()
                        << "\n- Resolved global pose: \n" << globalNodePose->toEigen()
                           ;

                nav::loc::arondto::Location data;
                data.globalRobotPose = globalNodePose->toEigen();

                armem::EntityUpdate& update = commit.add();
                update.entityID = getLocationProviderSegmentID().withEntityName(name);
                update.timeCreated = time;
                update.instancesData = { data.toAron() };
            }
        }

        if (not tab.dryRun.getValue())
        {
            armem::CommitResult result = proxies.navigationWriter.commit(commit);
            if (result.allSuccess())
            {
                ARMARX_IMPORTANT << "Successfully exported " << result.results.size() << " locations.";
            }
            else
            {
                ARMARX_WARNING << result.allErrorMessages();
            }
        }
        else
        {
            ARMARX_VERBOSE << "Dry Run - skipping commit.";
        }
    }


    void GraphImportExport::drawScene(const std::string& sceneName)
    {
        memoryx::GraphNodeBaseList graphNodes = proxies.graphSegment->getNodesByScene(sceneName);
        for (memoryx::GraphNodeBasePtr& node : graphNodes)
        {
            ARMARX_CHECK_NOT_NULL(node);

            armarx::FramedPosePtr pose = armarx::FramedPosePtr::dynamicCast(node->getPose());
            if (pose and not node->isMetaEntity())
            {
                const std::string nodeId = node->getId();

                const std::string name = node->getName();
                FramedPosePtr globalNodePose = FramedPosePtr::dynamicCast(proxies.graphNodePoseResolver->resolveToGlobalPose(node));
                ARMARX_CHECK_NOT_NULL(globalNodePose);
            }
        }

#if 0
        // Add edges
        for (memoryx::GraphNodeBasePtr& node : graphNodes)
        {
            std::string nodeId = node->getId();
            for (int i = 0; i < node->getOutdegree(); i++)
            {
                auto adjacent = memoryx::GraphNodeBasePtr::dynamicCast(node->getAdjacentNode(i)->getEntity());
                ARMARX_CHECK_NOT_NULL(adjacent);
                auto adjacentId = adjacent->getId();

                addEdge(nodeId, adjacentId);
            }
        }
#endif
    }


    void GraphImportExport::addNode(const memoryx::GraphNodeBasePtr& node)
    {
        (void) node;

    }

    armem::MemoryID GraphImportExport::getLocationProviderSegmentID()
    {
        return armem::MemoryID(properties.memoryName,
                               properties.locationCoreSegmentName,
                               tab.providerSegmentLine.getValue());
    }

    armem::MemoryID GraphImportExport::getGraphProviderSegmentID()
    {
        return armem::MemoryID(properties.memoryName,
                               properties.graphCoreSegmentName,
                               tab.providerSegmentLine.getValue());
    }


#if 0
    void GraphImportExport::addEdge(const std::string& node1Id, const std::string& node2Id)
    {
        if (!hasNode(node1Id))
        {
            ARMARX_WARNING << "Edge: " << node1Id << ", " << node2Id << " can't be created! Node " << node1Id << " does not exist.";
            return;
        }

        if (!hasNode(node2Id))
        {
            ARMARX_WARNING << "Edge: " << node1Id << ", " << node2Id << " can't be created! Node " << node2Id << " does not exist.";
            return;
        }

        auto node1dat = nodes.at(node1Id);
        auto node2dat = nodes.at(node2Id);

        if (hasEdge(node1Id, node2Id))
        {
            //nothing needs to be updated
            ARMARX_WARNING << "Edge: '" << node1dat.node->getName() << "' -> '" << node2dat.node->getName() << "' already exists.";
            return;
        }

        auto edgeId = toEdge(node1Id, node2Id);

        //add
        //table
        int row = ui.tableWidgetEdges->rowCount();
        ui.tableWidgetEdges->setRowCount(row + 1);
        ui.tableWidgetEdges->setItem(row, 0, new QTableWidgetItem {QString::fromStdString(node1dat.node->getName())});
        ui.tableWidgetEdges->setItem(row, 1, new QTableWidgetItem {QString::fromStdString(node2dat.node->getName())});
        //debug layer will be done later
        //scene
        QGraphicsLineItem* graphicsItem = dynamic_cast<QGraphicsLineItem*>(new GraphVisualizerGraphicsLineItem
        {
            *this, edgeId,
            node1dat.pose->position->x, -node1dat.pose->position->y,
            node2dat.pose->position->x, -node2dat.pose->position->y
        });
        //auto graphicsItem= scene->addLine(node1dat.pos->x,-node1dat.pos->y,
        //                                  node2dat.pos->x,-node2dat.pos->y);
        scene->addItem(graphicsItem);
        //setToolTip on graphicsItem does not work
        dynamic_cast<QGraphicsItem*>(graphicsItem)->setToolTip(QString {"Edge:"} +QString::fromStdString(node1dat.node->getName()) +
                QString {" <-> "} +QString::fromStdString(node2dat.node->getName()));
        //data
        EdgeData data {graphicsItem, row, false};
        edges[edgeId] = data;

        updateEdge(edgeId);
    }
#endif


}
